##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
Helper functions to generate complex data type context (cdt_ctx) objects for use with operations on nested CDTs (list, map, etc).

Example::

    import aerospike
    from aerospike import exception as ex
    from aerospike_helpers import cdt_ctx
    from aerospike_helpers.operations import map_operations
    from aerospike_helpers.operations import list_operations
    import sys

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}

    # Create a client and connect it to the cluster.
    try:
        client = aerospike.client(config).connect()
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    key = ("test", "demo", "foo")
    nested_list = [{"name": "John", "id": 100}, {"name": "Bill", "id": 200}]
    nested_list_bin_name = "nested_list"

    # Write the record.
    try:
        client.put(key, {nested_list_bin_name: nested_list})
    except ex.RecordError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))

    # EXAMPLE 1: read a value from the map nested at list index 1.
    try:
        ctx = [cdt_ctx.cdt_ctx_list_index(1)]

        ops = [
            map_operations.map_get_by_key(
                nested_list_bin_name, "id", aerospike.MAP_RETURN_VALUE, ctx
            )
        ]

        _, _, result = client.operate(key, ops)
        print("EXAMPLE 1, id is: ", result)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    # EXAMPLE 2: write a new nested map at list index 2 and get the value at its 'name' key.
    # NOTE: The map is appened to the list, then the value is read using the ctx.
    try:
        new_map = {"name": "Cindy", "id": 300}

        ctx = [cdt_ctx.cdt_ctx_list_index(2)]

        ops = [
            list_operations.list_append(nested_list_bin_name, new_map),
            map_operations.map_get_by_key(
                nested_list_bin_name, "name", aerospike.MAP_RETURN_VALUE, ctx
            ),
        ]

        _, _, result = client.operate(key, ops)
        print("EXAMPLE 2, name is: ", result)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    # Cleanup and close the connection to the Aerospike cluster.
    client.remove(key)
    client.close()

    """
    EXPECTED OUTPUT:
    EXAMPLE 1, id is:  {'nested_list': 200}
    EXAMPLE 2, name is:  {'nested_list': 'Cindy'}
    """
'''
import aerospike


CDT_CTX_ORDER_KEY = "order_key"
CDT_CTX_PAD_KEY = "pad_key"

class _cdt_ctx:
    """
    Class used to represent a single ctx_operation.
    """
    def __init__(self, *, id=None, value=None, extra_args=None):
        self.id = id
        self.value = value
        self.extra_args = extra_args


def cdt_ctx_list_index(index):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a list by index.
    If the index is negative, the lookup starts backwards from the end of the list.
    If it is out of bounds, a parameter error will be returned.

    Args:
        index (int): The index to look for in the list.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_INDEX, value=index)


def cdt_ctx_list_rank(rank):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a list by rank.
    If the rank is negative, the lookup starts backwards from the largest rank value.

    Args:
        rank (int): The rank to look for in the list.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_RANK, value=rank)


def cdt_ctx_list_value(value):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a list by value.

    Args:
        value (object): The value to look for in the list.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_VALUE, value=value)


def cdt_ctx_list_index_create(index: int, order: int = 0, pad: bool = False) -> _cdt_ctx:
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    Create a list with the given sort order at the given index.

    Args:
        key (object): The index to create the list at.
        order (int): The :ref:`sort order <aerospike_list_order>` to create the List with (default ``aerospike.LIST_UNORDERED``)
        pad (bool): If index is out of bounds and pad is True, the list will be created at index and empty list elements inserted behind it.
            Pad is only compatible with unordered lists.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_INDEX_CREATE, value=index, extra_args={CDT_CTX_ORDER_KEY: order, CDT_CTX_PAD_KEY: pad})


def cdt_ctx_map_index(index):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a map by index.
    If the index is negative, the lookup starts backwards from the end of the map.
    If it is out of bounds, a parameter error will be returned.

    Args:
        index (int): The index to look for in the map.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_INDEX, value=index)


def cdt_ctx_map_rank(rank):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a map by index.
    If the rank is negative, the lookup starts backwards from the largest rank value.

    Args:
        rank (int): The rank to look for in the map.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_RANK, value=rank)


def cdt_ctx_map_key(key):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a map by key.

    Args:
        key (object): The key to look for in the map.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_KEY, value=key)


def cdt_ctx_map_value(value):
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    The cdt_ctx object is initialized to lookup an object in a map by value.

    Args:
        value (object): The value to look for in the map.
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_VALUE, value=value)


def cdt_ctx_map_key_create(key: any, order: int = 0) -> _cdt_ctx:
    """Creates a nested cdt_ctx object for use with list or map operations.
    
    Create a map with the given sort order at the given key.

    Args:
        key (object): The key to create the map at.
        order (int): The :ref:`sort order <aerospike_map_order>` to create the List with (default ``aerospike.MAP_UNORDERED``)
    
    Returns:
        A cdt_ctx object, a list of these is usable with list and map operations.
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_KEY_CREATE, value=key, extra_args={CDT_CTX_ORDER_KEY: order})



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
This module provides helper functions to produce dictionaries to be used with the
:mod:`aerospike.Client.operate` and :mod:`aerospike.Client.operate_ordered` methods of the aerospike module.

Expression operations support reading and writing the result of Aerospike expressions.

.. note:: Requires server version >= 5.6.0

"""
import aerospike
from aerospike_helpers.expressions import resources


OP_KEY = "op"
BIN_KEY = "bin"
EXPR_KEY = "expr"
EXPR_FLAGS_KEY = "expr_flags"


def expression_read(bin_name: str, expression: resources._BaseExpr, expression_read_flags: int=0):
    """Create an expression read operation dictionary.

    Reads and returns the value produced by the evaluated expression.

    Args:
        bin_name (str): The name of the bin to read from. Even if no bin is being read from, the value will be returned with this bin name.
        expression: A compiled Aerospike expression, see :ref:`aerospike_operation_helpers.expressions`.
        expression_read_flags (int): :ref:`aerospike_expression_read_flags` (default ``aerospike.EXP_READ_DEFAULT``)
    Returns:
        A dictionary to be passed to operate or operate_ordered.

    Example::

        # Read the value of int bin "balance".
        # Let 'client' be a connected aerospike client.
        # Let int bin 'balance' == 50.

        from aerospike_helpers.operations import expression_operations as expressions
        from aerospike_helpers.expressions import *
        
        expr = IntBin("balance").compile()
        ops = [
            expressions.expression_read("balance", expr)
        ]
        _, _, res = client.operate(self.key, ops)
       print(res)

       # EXPECTED OUTPUT: {"balance": 50}
    """

    op_dict = {
        OP_KEY: aerospike.OP_EXPR_READ,
        BIN_KEY: bin_name,
        EXPR_KEY: expression,
        EXPR_FLAGS_KEY: expression_read_flags
    }

    return op_dict


def expression_write(bin_name: str, expression: resources._BaseExpr, expression_write_flags: int=0):
    """Create an expression write operation dictionary.

    Writes the value produced by the evaluated expression to the supplied bin.

    Args:
        bin_name (str): The name of the bin to write to.
        expression: A compiled Aerospike expression, see :ref:`aerospike_operation_helpers.expressions`.
        expression_write_flags (int): :ref:`aerospike_expression_write_flags` such as ``aerospike.EXP_WRITE_UPDATE_ONLY | aerospike.EXP_WRITE_POLICY_NO_FAIL``   (default ``aerospike.EXP_WRITE_DEFAULT``).
    Returns:
        A dictionary to be passed to operate or operate_ordered.

    Example::

        # Write the value of int bin "balance" + 50 back to "balance".
        # Let 'client' be a connected aerospike client.
        # Let int bin 'balance' == 50.

        from aerospike_helpers.operations import expression_operations as expressions
        from aerospike_helpers.expressions import *
        
        expr = Add(IntBin("balance"), 50).compile()
        ops = [
            expressions.expression_write("balance", expr)
        ]
        client.operate(self.key, ops)
        _, _, res = client.get(self.key)
       print(res)

       # EXPECTED OUTPUT: {"balance": 100}
    """

    op_dict = {
        OP_KEY: aerospike.OP_EXPR_WRITE,
        BIN_KEY: bin_name,
        EXPR_KEY: expression,
        EXPR_FLAGS_KEY: expression_write_flags
    }

    return op_dict



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
List expressions contain expressions for reading and modifying Lists. Most of
these operations are from the standard :mod:`List API <aerospike_helpers.operations.list_operations>`.

Example::

    import aerospike_helpers.expressions as exp
    #Take the size of list bin "a".
    expr = exp.ListSize(None, exp.ListBin("a")).compile()
'''

#from __future__ import annotations
from itertools import chain
from typing import List, Optional, Tuple, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import ResultType
from aerospike_helpers.expressions.resources import _Keys
from aerospike_helpers.expressions.base import ListBin

######################
# List Mod Expressions
######################

TypeBinName = Union[_BaseExpr, str]
TypeListValue = Union[_BaseExpr, List[Any]]
TypeIndex = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeCTX = Union[None, List[cdt_ctx._cdt_ctx]]
TypeRank = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeCount = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeValue = Union[_BaseExpr, Any]
TypePolicy = Union[Dict[str, Any], None]


class ListAppend(_BaseExpr):
    """Create an expression that appends value to end of list."""
    _op = aerospike.OP_LIST_APPEND

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                value (TypeValue): Value or value expression to append to list.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Check if length of list bin "a" is > 5 after appending 1 item.
                expr = exp.GT(
                        exp.ListSize(None, exp.ListAppend(None, None, 3, exp.ListBin("a"))),
                        5).compile()
        """
        self._children = (
            value,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_CRMOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListAppendItems(_BaseExpr):
    """Create an expression that appends a list of items to the end of a list."""
    _op = aerospike.OP_LIST_APPEND_ITEMS

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                value (TypeValue): List or list expression of items to be appended.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Check if length of list bin "a" is > 5 after appending multiple items.
                expr = exp.GT(
                        exp.ListSize(None, exp.ListAppendItems(None, None, [3, 2], exp.ListBin("a"))),
                        5).compile()
        """
        self._children = (
            value,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_CRMOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListInsert(_BaseExpr):
    """Create an expression that inserts value to specified index of list."""
    _op = aerospike.OP_LIST_INSERT

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', index: 'TypeIndex', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                index (TypeIndex): Target index for insertion, integer or integer expression.
                value (TypeValue): Value or value expression to be inserted.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Check if list bin "a" has length > 5 after insert.
                expr = exp.GT(
                        exp.ListSize(None, exp.ListInsert(None, None, 0, 3, exp.ListBin("a"))),
                        5).compile()
        """
        self._children = (
            index,
            value,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_MOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListInsertItems(_BaseExpr):
    """Create an expression that inserts each input list item starting at specified index of list."""
    _op = aerospike.OP_LIST_INSERT_ITEMS

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', index: 'TypeIndex', values: 'TypeListValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                index (TypeIndex): Target index where item insertion will begin, integer or integer expression.
                values (TypeListValue): List or list expression of items to be inserted.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Check if list bin "a" has length > 5 after inserting items.
                expr = exp.GT(
                        exp.ListSize(None, exp.ListInsertItems(None, None, 0, [4, 7], exp.ListBin("a"))),
                        5).compile()
        """
        self._children = (
            index,
            values,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_MOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListIncrement(_BaseExpr):
    """Create an expression that increments list[index] by value."""
    _op = aerospike.OP_LIST_INCREMENT

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', index: 'TypeIndex', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                index (TypeIndex): Index of value to increment.
                value (TypeValue): Value or value expression.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Check if incremented value in list bin "a" is the largest in the list.
                expr = exp.Eq(
                        exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, -1, #rank of -1 == largest element.
                            exp.ListIncrement(None, None, 1, 5, exp.ListBin("a"))),
                        exp.ListGetByIndex(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, 1,
                            exp.ListIncrement(None, None, 1, 5, exp.ListBin("a")))
                ).compile()
        """
        self._children = (
            index,
            value,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_CRMOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListSet(_BaseExpr):
    """Create an expression that sets item value at specified index in list."""
    _op = aerospike.OP_LIST_SET

    def __init__(self, ctx: 'TypeCTX', policy: 'TypePolicy', index: 'TypeIndex', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                policy (TypePolicy): Optional dictionary of :ref:`List policies <aerospike_list_policies>`.
                index (TypeIndex): index of value to set.
                value (TypeValue): value or value expression to set index in list to.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Get smallest element in list bin "a" after setting index 1 to 10.
                expr = exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, 0,
                                exp.ListSet(None, None, 1, 10, exp.ListBin("a"))).compile()
        """
        self._children = (
            index,
            value,
            _GenericExpr(_ExprOp._AS_EXP_CODE_CDT_LIST_MOD, 0, {_Keys.LIST_POLICY_KEY: policy} if policy is not None else {}),
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.LIST_POLICY_KEY] = policy


class ListClear(_BaseExpr):
    """Create an expression that removes all items in a list."""
    _op = aerospike.OP_LIST_CLEAR

    def __init__(self, ctx: 'TypeCTX', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: List expression.

            Example::

                # Clear list value of list nested in list bin "a" index 1.
                from aerospike_helpers import cdt_ctx
                expr = exp.ListClear([cdt_ctx.cdt_ctx_list_index(1)], "a").compile()
        """
        self._children = (
            bin if isinstance(bin, _BaseExpr) else ListBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListSort(_BaseExpr):
    """Create an expression that sorts a list."""
    _op = aerospike.OP_LIST_SORT

    def __init__(self, ctx: 'TypeCTX', order: int, bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                order (int): Optional flags modifiying the behavior of list_sort. This should be constructed by bitwise or'ing together values from :ref:`aerospike_list_sort_flag`.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Get value of sorted list bin "a".
                expr = exp.ListSort(None, aerospike.LIST_SORT_DEFAULT, "a").compile()
        """
        self._children = (
            bin if isinstance(bin, _BaseExpr) else ListBin(bin),
        )
        self._fixed = {_Keys.LIST_ORDER_KEY: order}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByValue(_BaseExpr):
    """Create an expression that removes list items identified by value."""
    _op = aerospike.OP_LIST_REMOVE_BY_VALUE

    def __init__(self, ctx: 'TypeCTX', value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                value (TypeValue): Value or value expression to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # See if list bin "a", with `3` removed, is equal to list bin "b".
                expr = exp.Eq(exp.ListRemoveByValue(None, 3, exp.ListBin("a")), ListBin("b")).compile()
        """
        self._children = (
            value,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByValueList(_BaseExpr):
    """Create an expression that removes list items identified by values."""
    _op = aerospike.OP_LIST_REMOVE_BY_VALUE_LIST

    def __init__(self, ctx: 'TypeCTX', values: 'TypeListValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                values (TypeListValue): List of values or list expression.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove elements with values [1, 2, 3] from list bin "a".
                expr = exp.ListRemoveByValueList(None, [1, 2, 3], exp.ListBin("a")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByValueRange(_BaseExpr):
    """ Create an expression that removes list items identified by value range
        (begin inclusive, end exclusive). If begin is None, the range is less than end.
        If end is None, the range is greater than or equal to begin.
    """
    _op = aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE

    def __init__(self, ctx: 'TypeCTX', begin: 'TypeValue', end: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                begin (TypeValue): Begin value or value expression for range.
                end (TypeValue): End value or value expression for range.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove list of items with values >= 3 and < 7 from list bin "a".
                expr = exp.ListRemoveByValueRange(None, 3, 7, exp.ListBin("a")).compile()
        """
        self._children = (
            begin,
            end,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByValueRelRankToEnd(_BaseExpr):
    """Create an expression that removes list items nearest to value and greater by relative rank."""
    _op = aerospike.OP_LIST_REMOVE_BY_REL_RANK_RANGE_TO_END

    def __init__(self, ctx: 'TypeCTX', value: 'TypeValue', rank: 'TypeRank', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                value (TypeValue): Start value or value expression.
                rank (TypeRank): Rank integer or integer expression.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove elements larger than 4 by relative rank in list bin "a".
                expr = exp.ListRemoveByValueRelRankToEnd(None, 4, 1, exp.ListBin("a")).compile()
        """
        self._children = (
            value,
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByValueRelRankRange(_BaseExpr):
    """ Create an expression that removes list items nearest to value and greater by relative rank with a
        count limit.
    """
    _op = aerospike.OP_LIST_REMOVE_BY_REL_RANK_RANGE

    def __init__(self, ctx: 'TypeCTX', value: 'TypeValue', rank: 'TypeRank', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                value (TypeValue): Start value or value expression.
                rank (TypeRank): Rank integer or integer expression.
                count (TypeCount): How many elements to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # After removing the 3 elements larger than 4 by relative rank, does list bin "a" include 9?.
                expr = exp.GT(
                        exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, 9,
                            exp.ListRemoveByValueRelRankRange(None, 4, 1, 0, exp.ListBin("a"))),
                        0).compile()
        """
        self._children = (
            value,
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByIndex(_BaseExpr):
    """Create an expression that removes "count" list items starting at specified index."""
    _op = aerospike.OP_LIST_REMOVE_BY_INDEX

    def __init__(self, ctx: 'TypeCTX', index: 'TypeIndex', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                index (TypeIndex): Index integer or integer expression of element to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Get size of list bin "a" after index 3 has been removed.
                expr = exp.ListSize(None, exp.ListRemoveByIndex(None, 3, exp.ListBin("a"))).compile()
        """
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByIndexRangeToEnd(_BaseExpr):
    """Create an expression that removes list items starting at specified index to the end of list."""
    _op = aerospike.OP_LIST_REMOVE_BY_INDEX_RANGE_TO_END

    def __init__(self, ctx: 'TypeCTX', index: 'TypeIndex', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                index (TypeIndex): Starting index integer or integer expression of elements to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove all elements starting from index 3 in list bin "a".
                expr = exp.ListRemoveByIndexRangeToEnd(None, 3, exp.ListBin("a")).compile()
        """
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByIndexRange(_BaseExpr):
    """Create an expression that removes "count" list items starting at specified index."""
    _op = aerospike.OP_LIST_REMOVE_BY_INDEX_RANGE

    def __init__(self, ctx: 'TypeCTX', index: 'TypeIndex', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                index (TypeIndex): Starting index integer or integer expression of elements to remove.
                count (TypeCount): Integer or integer expression, how many elements to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Get size of list bin "a" after index 3, 4, and 5 have been removed.
                expr = exp.ListSize(None, exp.ListRemoveByIndexRange(None, 3, 3, exp.ListBin("a"))).compile()
        """
        self._children = (
            index,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByRank(_BaseExpr):
    """Create an expression that removes list item identified by rank."""
    _op = aerospike.OP_LIST_REMOVE_BY_RANK

    def __init__(self, ctx: 'TypeCTX', rank: 'TypeRank', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                rank (TypeRank): Rank integer or integer expression of element to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove smallest value in list bin "a".
                expr = exp.ListRemoveByRank(None, 0, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByRankRangeToEnd(_BaseExpr):
    """Create an expression that removes list items starting at specified rank to the last ranked item."""
    _op = aerospike.OP_LIST_REMOVE_BY_RANK_RANGE_TO_END

    def __init__(self, ctx: 'TypeCTX', rank: 'TypeRank', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                rank (TypeRank): Rank integer or integer expression of element to start removing at.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove the 2 largest elements from List bin "a".
                expr = exp.ListRemoveByRankRangeToEnd(None, -2, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListRemoveByRankRange(_BaseExpr):
    """Create an expression that removes "count" list items starting at specified rank."""
    _op = aerospike.OP_LIST_REMOVE_BY_RANK_RANGE

    def __init__(self, ctx: 'TypeCTX', rank: 'TypeRank', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                rank (TypeRank): Rank integer or integer expression of element to start removing at.
                count (TypeCount): Count integer or integer expression of elements to remove.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: list expression.

            Example::

                # Remove the 3 smallest items from list bin "a".
                expr = exp.ListRemoveByRankRange(None, 0, 3, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


#######################
# List Read Expressions
#######################


class ListSize(_BaseExpr):
    """Create an expression that returns list size."""
    _op = aerospike.OP_LIST_SIZE

    def __init__(self, ctx: 'TypeCTX', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Integer expression.

            Example::

                #Take the size of list bin "a".
                expr = exp.ListSize(None, exp.ListBin("a")).compile()
        """
        self._children = (
            bin if isinstance(bin, _BaseExpr) else ListBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByValue(_BaseExpr):
    """ Create an expression that selects list items identified by value and returns selected
        data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_VALUE

    def __init__(self, ctx: 'TypeCTX', return_type: int, value: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value (TypeValue): Value or value expression of element to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the index of the element with value, 3, in list bin "a".
                expr = exp.ListGetByValue(None, aerospike.LIST_RETURN_INDEX, 3, exp.ListBin("a")).compile()
        """
        self._children = (
            value,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByValueRange(_BaseExpr):
    """ Create an expression that selects list items identified by value range and returns selected
        data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_VALUE_RANGE

    def __init__(
        self,
        ctx: 'TypeCTX',
        return_type: int,
        value_begin: 'TypeValue',
        value_end: 'TypeValue',
        bin: 'TypeBinName'
    ):
        """ Create an expression that selects list items identified by value range and returns selected
            data specified by return_type.

            Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value_begin (TypeValue): Value or value expression of first element to get.
                value_end (TypeValue): Value or value expression of ending element.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get rank of values between 3 (inclusive) and 7 (exclusive) in list bin "a".
                expr = exp.ListGetByValueRange(None, aerospike.LIST_RETURN_RANK, 3, 7, exp.ListBin("a")).compile()
        """
        self._children = (
            value_begin,
            value_end,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByValueList(_BaseExpr):
    """ Create an expression that selects list items identified by values and returns selected
        data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_VALUE_LIST

    def __init__(self, ctx: 'TypeCTX', return_type: int, value: 'TypeListValue', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values  Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value (TypeListValue): List or list expression of values of elements to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                #Get the indexes of the the elements in list bin "a" with values [3, 6, 12].
                expr = exp.ListGetByValueList(None, aerospike.LIST_RETURN_INDEX, [3, 6, 12], exp.ListBin("a")).compile()
        """
        self._children = (
            value,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByValueRelRankRangeToEnd(_BaseExpr):
    """Create an expression that selects list items nearest to value and greater by relative rank"""
    _op = aerospike.OP_LIST_GET_BY_VALUE_RANK_RANGE_REL_TO_END

    def __init__(self, ctx: 'TypeCTX', return_type: int, value: 'TypeValue', rank: 'TypeRank', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value (TypeValue): Value or vaule expression to get items relative to.
                rank (TypeRank): Rank intger expression. rank relative to "value" to start getting elements.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the values of all elements in list bin "a" larger than 3.
                expr = exp.ListGetByValueRelRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 3, 1, exp.ListBin("a")).compile()
        """
        self._children = (
            value,
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByValueRelRankRange(_BaseExpr):
    """ Create an expression that selects list items nearest to value and greater by relative rank with a
        count limit and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_VALUE_RANK_RANGE_REL

    def __init__(self, ctx: 'TypeCTX', return_type: int, value: 'TypeValue', rank: 'TypeRank', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value (TypeValue): Value or vaule expression to get items relative to.
                rank (TypeRank): Rank intger expression. rank relative to "value" to start getting elements.
                count (TypeCount): Integer value or integer value expression, how many elements to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the next 2 values in list bin "a" larger than 3.
                expr = exp.ListGetByValueRelRankRange(None, aerospike.LIST_RETURN_VALUE, 3, 1, 2, exp.ListBin("a")).compile()
        """
        self._children = (
            value,
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByIndex(_BaseExpr):
    """ Create an expression that selects list item identified by index
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_INDEX

    def __init__(
        self,
        ctx: 'TypeCTX',
        return_type: int,
        value_type: int,
        index: 'TypeIndex',
        bin: 'TypeBinName',
    ):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values 
                value_type (int): The value type that will be returned by this expression (ResultType).
                index (TypeIndex): Integer or integer expression of index to get element at.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the value at index 0 in list bin "a". (assume this value is an integer)
                expr = exp.ListGetByIndex(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, 0, exp.ListBin("a")).compile()
        """    
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.VALUE_TYPE_KEY: value_type, _Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByIndexRangeToEnd(_BaseExpr):
    """ Create an expression that selects list items starting at specified index to the end of list
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_INDEX_RANGE_TO_END

    def __init__(self, ctx: 'TypeCTX', return_type: int, index: 'TypeIndex', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                index (TypeIndex): Integer or integer expression of index to start getting elements at.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get element 5 to end from list bin "a".
                expr = exp.ListGetByIndexRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 5, exp.ListBin("a")).compile()
        """
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByIndexRange(_BaseExpr):
    """ Create an expression that selects "count" list items starting at specified index
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_INDEX_RANGE

    def __init__(self, ctx: 'TypeCTX', return_type: int, index: 'TypeIndex', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                index (TypeIndex): Integer or integer expression of index to start getting elements at.
                count (TypeCount): Integer or integer expression for count of elements to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get elements at indexes 3, 4, 5, 6 in list bin "a".
                expr = exp.ListGetByIndexRange(None, aerospike.LIST_RETURN_VALUE, 3, 4, exp.ListBin("a")).compile()
        """
        self._children = (
            index,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByRank(_BaseExpr):
    """ Create an expression that selects list item identified by rank
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_RANK

    def __init__(
        self,
        ctx: 'TypeCTX',
        return_type: int,
        value_type: int,
        rank: 'TypeRank',
        bin: 'TypeBinName',
    ):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                value_type (int): The value type that will be returned by this expression (ResultType).
                rank (TypeRank): Rank integer or integer expression of element to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the smallest element in list bin "a".
                expr = exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, aerospike.ResultType.INTEGER, 0, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.VALUE_TYPE_KEY: value_type, _Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByRankRangeToEnd(_BaseExpr):
    """ Create an expression that selects list items starting at specified rank to the last ranked item
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_RANK_RANGE_TO_END

    def __init__(self, ctx: 'TypeCTX', return_type: int, rank: 'TypeRank', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                rank (TypeRank): Rank integer or integer expression of first element to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the three largest elements in list bin "a".
                expr = exp.ListGetByRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, -3, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class ListGetByRankRange(_BaseExpr):
    """ Create an expression that selects "count" list items starting at specified rank
        and returns selected data specified by return_type.
    """
    _op = aerospike.OP_LIST_GET_BY_RANK_RANGE

    def __init__(self, ctx: 'TypeCTX', return_type: int, rank: 'TypeRank', count: 'TypeCount', bin: 'TypeBinName'):
        """ Args:
                ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation objects.
                return_type (int): Value specifying what should be returned from the operation.
                    This should be one of the :ref:`list_return_types` values.
                rank (TypeRank): Rank integer or integer expression of first element to get.
                count (TypeCount): Count integer or integer expression for how many elements to get.
                bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or :class:`~aerospike_helpers.expressions.base.ListBin`.

            :return: Expression.

            Example::

                # Get the 3 smallest elements in list bin "a".
                expr = exp.ListGetByRankRange(None, aerospike.LIST_RETURN_VALUE, 0, 3, exp.ListBin("a")).compile()
        """
        self._children = (
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else ListBin(bin)
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
Helper functions to create bit operation dictionary arguments for
the :meth:`aerospike.operate` and :meth:`aerospike.operate_ordered` methods of the aerospike client.

    .. note:: Bitwise operations require server version >= 4.6.0

Bit offsets are oriented left to right. Negative offsets are supported and start backwards from the end of the target bitmap.

Offset examples:
    * 0: leftmost bit in the map
    * 4: fifth bit in the map
    * -1: rightmost bit in the map
    * -4: 3 bits from rightmost

Example::

    import aerospike
    from aerospike import exception as ex
    from aerospike_helpers.operations import bitwise_operations
    import sys

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}

    # Create a client and connect it to the cluster.
    try:
        client = aerospike.client(config).connect()
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    key = ("test", "demo", "foo")
    five_ones_bin_name = "bitwise1"
    five_one_blob = bytearray([1] * 5)

    # Write the record.
    try:
        client.put(key, {five_ones_bin_name: five_one_blob})
    except ex.RecordError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))

    # EXAMPLE 1: resize the five_ones bin to a bytesize of 10.
    try:
        ops = [bitwise_operations.bit_resize(five_ones_bin_name, 10)]

        _, _, bins = client.get(key)
        _, _, _ = client.operate(key, ops)
        _, _, newbins = client.get(key)
        print("EXAMPLE 1: before resize: ", bins)
        print("EXAMPLE 1: is now: ", newbins)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    # EXAMPLE 2: shrink the five_ones bin to a bytesize of 5 from the front.
    try:
        ops = [
            bitwise_operations.bit_resize(
                five_ones_bin_name, 5, resize_flags=aerospike.BIT_RESIZE_FROM_FRONT
            )
        ]

        _, _, bins = client.get(key)
        _, _, _ = client.operate(key, ops)
        _, _, newbins = client.get(key)
        print("EXAMPLE 2: before resize: ", bins)
        print("EXAMPLE 2: is now: ", newbins)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    # Cleanup and close the connection to the Aerospike cluster.
    client.remove(key)
    client.close()

    """
    EXPECTED OUTPUT:
    EXAMPLE 1: before resize:  {'bitwise1': bytearray(b'\\x01\\x01\\x01\\x01\\x01')}
    EXAMPLE 1: is now:  {'bitwise1': bytearray(b'\\x01\\x01\\x01\\x01\\x01\\x00\\x00\\x00\\x00\\x00')}
    EXAMPLE 2: before resize:  {'bitwise1': bytearray(b'\\x01\\x01\\x01\\x01\\x01\\x00\\x00\\x00\\x00\\x00')}
    EXAMPLE 2: is now:  {'bitwise1': bytearray(b'\\x00\\x00\\x00\\x00\\x00')}
    """

Example::

    import aerospike
    from aerospike import exception as e
    from aerospike_helpers.operations import bitwise_operations

    config = {'hosts': [('127.0.0.1', 3000)]}
    try:
        client = aerospike.client(config).connect()
    except e.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(2)

    key = ('test', 'demo', 'bit_example')
    five_one_blob = bytearray([1] * 5)
    five_one_bin = 'bitwise1'

    try:
        if client.exists(key):
            client.remove(key)
        bit_policy = {
            'map_write_mode': aerospike.BIT_WRITE_DEFAULT,
        }
        client.put(
            key,
            {
                five_one_bin: five_one_blob
            }
        )

        # Example 1: read bits
        ops = [
            bitwise_operations.bit_get(five_one_bin, 0, 40)
        ]
        print('=====EXAMPLE1=====')
        _, _, results = client.operate(key, ops)
        print(results)

        # Example 2: modify bits using the 'or' op, then read bits
        ops = [
            bitwise_operations.bit_or(five_one_bin, 0, 8, 1, bytearray([255]), bit_policy),
            bitwise_operations.bit_get(five_one_bin, 0, 40)
        ]
        print('=====EXAMPLE2=====')
        _, _, results = client.operate(key, ops)
        print(results)

        # Example 3: modify bits using the 'remove' op, then read bits'
        ops = [
            bitwise_operations.bit_remove(five_one_bin, 0, 2, bit_policy),
            bitwise_operations.bit_get(five_one_bin, 0, 24)
        ]
        print('=====EXAMPLE3=====')
        _, _, results = client.operate(key, ops)
        print(results)

    except e.AerospikeError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))

    client.close()

    """
    EXPECTED OUTPUT:
    =====EXAMPLE1=====
    {'bitwise1': bytearray(b'\\x01\\x01\\x01\\x01\\x01')}
    =====EXAMPLE2=====
    {'bitwise1': bytearray(b'\\xff\\x01\\x01\\x01\\x01')}
    =====EXAMPLE3=====
    {'bitwise1': bytearray(b'\\x01\\x01\\x01')}
    """

.. seealso:: `Bits (Data Types) <https://www.aerospike.com/docs/guide/bitwise.html>`_.
'''
import aerospike

BIN_KEY = "bin"
BYTE_SIZE_KEY = "byte_size"
BYTE_OFFSET_KEY = "byte_offset"
BIT_OFFSET_KEY = "bit_offset"
BIT_SIZE_KEY = "bit_size"
VALUE_BYTE_SIZE_KEY = "value_byte_size"
VALUE_KEY = "value"
COUNT_KEY = "count"
INDEX_KEY = "index"
KEY_KEY = "key"
OFFSET_KEY = "offset"
OP_KEY = "op"
POLICY_KEY = "policy"
RESIZE_FLAGS_KEY = "resize_flags"
SIGN_KEY = "sign"
ACTION_KEY = "action"


def bit_resize(bin_name: str, byte_size, policy=None, resize_flags: int=0):
    """Creates a bit_resize_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Change the size of a bytes bin stored in a record on the Aerospike Server.

    Args:
        bin_name (str): The name of the bin containing the map.
        byte_size (int): The new size of the bytes.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.
        resize_flags (int): :ref:`aerospike_bitwise_resize_flag` modifying the resize behavior (default ``aerospike.BIT_RESIZE_DEFAULT``), such as ``aerospike.BIT_RESIZE_GROW_ONLY | aerospike.BIT_RESIZE_FROM_FRONT``.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_RESIZE,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        RESIZE_FLAGS_KEY: resize_flags,
        BYTE_SIZE_KEY: byte_size
    }


def bit_remove(bin_name: str, byte_offset, byte_size, policy=None):
    """Creates a bit_remove_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Remove bytes from bitmap at byte_offset for byte_size.

    Args:
        bin_name (str): The name of the bin containing the map.
        byte_offset (int): Position of bytes to be removed.
        byte_size (int): How many bytes to remove.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_REMOVE,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BYTE_OFFSET_KEY: byte_offset,
        BYTE_SIZE_KEY: byte_size
    }


def bit_set(bin_name: str, bit_offset, bit_size, value_byte_size, value, policy=None):
    """Creates a bit_set_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Set the value on a bitmap at bit_offset for bit_size in a record on the Aerospike Server.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will be set.
        bit_size (int): How many bits of value to write.
        value_byte_size (int): Size of value in bytes.
        value (bytes, bytearray): The value to be set.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_SET,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_BYTE_SIZE_KEY: value_byte_size,
        VALUE_KEY: value
    }


def bit_count(bin_name: str, bit_offset, bit_size):
    """Creates a bit_count_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns an integer count of all set bits starting at bit_offset for bit_size bits.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the set bits will begin being counted.
        bit_size (int): How many bits will be considered for counting.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_COUNT,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size
    }


def bit_add(bin_name: str, bit_offset, bit_size, value, sign, action, policy=None):
    """Creates a bit_add_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Creates a bit add operation. Server adds value to the bin at bit_offset for bit_size.
    bit_size must <= 64. If Sign is true value will be treated as a signed number.
    If an underflow or overflow occurs, as_bit_overflow_action is used. Server returns nothing.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will be added.
        bit_size (int): How many bits of value to add.
        value (int): The value to be added.
        sign (bool): True: treat value as signed, False: treat value as unsigned.
        action (aerospike.constant): Action taken if an overflow/underflow occurs.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_ADD,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: value,
        SIGN_KEY: sign,
        ACTION_KEY: action
    }


def bit_and(bin_name: str, bit_offset, bit_size, value_byte_size, value, policy=None):
    """Creates a bit_and_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Creates a bit and operation. Server performs an and op with value and bitmap in bin
    at bit_offset for bit_size. Server returns nothing.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will be modified.
        bit_size (int): How many bits of value to and.
        value_byte_size (int): Length of value in bytes.
        value (bytes, bytearray): Bytes to be used in and operation.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_AND,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_BYTE_SIZE_KEY: value_byte_size,
        VALUE_KEY: value
    }


def bit_get(bin_name: str, bit_offset, bit_size):
    """Creates a bit_get_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns bits from bitmap starting at bit_offset for bit_size.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being read.
        bit_size (int): How many bits to get.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_GET,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size
    }


def bit_get_int(bin_name: str, bit_offset, bit_size, sign):
    """Creates a bit_get_int_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns an integer formed from the bits read
    from bitmap starting at bit_offset for bit_size.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being read.
        bit_size (int): How many bits to get.
        sign (bool): True: Treat read value as signed. False: treat read value as unsigned.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_GET_INT,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        SIGN_KEY: sign
    }


def bit_insert(bin_name: str, byte_offset, value_byte_size, value, policy=None):
    """Creates a bit_insert_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server inserts the bytes from value into the bitmap at byte_offset.
    No value is returned.

    Args:
        bin_name (str): The name of the bin containing the map.
        byte_offset (int): The offset where the bytes will be inserted.
        value_byte_size (int): Size of value in bytes.
        value (bytes, bytearray): The value to be inserted.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.


    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_INSERT,
        BIN_KEY: bin_name,
        BYTE_OFFSET_KEY: byte_offset,
        VALUE_BYTE_SIZE_KEY: value_byte_size,
        VALUE_KEY: value,
        POLICY_KEY: policy
    }


def bit_lscan(bin_name: str, bit_offset, bit_size, value):
    """Creates a bit_lscan_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns an integer representing the bit offset of the first occurence
    of the specified value bit. Starts scanning at bit_offset for bit_size. Returns
    -1 if value not found.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being scanned.
        bit_size (int): How many bits to scan.
        value (bool): True: look for 1, False: look for 0.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_LSCAN,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: value
    }


def bit_lshift(bin_name: str, bit_offset, bit_size, shift, policy=None):
    """Creates a bit_lshift_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server left shifts bitmap starting at bit_offset for bit_size by shift bits.
    No value is returned.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being shifted.
        bit_size (int): The number of bits that will be shifted by shift places.
        shift (int): How many bits to shift by.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_LSHIFT,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: shift,
        POLICY_KEY: policy
    }


def bit_not(bin_name: str, bit_offset, bit_size, policy=None):
    """Creates a bit_not_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server negates bitmap starting at bit_offset for bit_size.
    No value is returned.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being scanned.
        bit_size (int): How many bits to scan.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_NOT,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        POLICY_KEY: policy
    }


def bit_or(bin_name: str, bit_offset, bit_size, value_byte_size, value, policy=None):
    """Creates a bit_or_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Creates a bit or operation. Server performs bitwise or with value and bitmap in bin
    at bit_offset for bit_size. Server returns nothing.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being compared.
        bit_size (int): How many bits of value to or.
        value_byte_size (int): Length of value in bytes.
        value (bytes/byte array): Value to be used in or operation.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_OR,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_BYTE_SIZE_KEY: value_byte_size,
        VALUE_KEY: value
    }


def bit_rscan(bin_name: str, bit_offset, bit_size, value):
    """Creates a bit_rscan_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns an integer representing the bit offset of the last occurence
    of the specified value bit. Starts scanning at bit_offset for bit_size. Returns
    -1 if value not found.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being scanned.
        bit_size (int): How many bits to scan.
        value (bool): True: Look for 1, False: look for 0.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_RSCAN,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: value
    }


def bit_rshift(bin_name: str, bit_offset, bit_size, shift, policy=None):
    """Creates a bit_rshift_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server right shifts bitmap starting at bit_offset for bit_size by shift bits.
    No value is returned.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being shifted.
        bit_size (int): The number of bits that will be shifted by shift places.
        shift (int): How many bits to shift by.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_RSHIFT,
        BIN_KEY: bin_name,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: shift,
        POLICY_KEY: policy
    }


def bit_subtract(bin_name: str, bit_offset, bit_size, value, sign, action, policy=None):
    """Creates a bit_subtract_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Creates a bit add operation. Server subtracts value from the bits at bit_offset for bit_size.
    bit_size must <= 64. If sign is true value will be treated as a signed number.
    If an underflow or overflow occurs, as_bit_overflow_action is used. Server returns nothing.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will be subtracted.
        bit_size (int): How many bits of value to subtract.
        value (int): The value to be subtracted.
        sign (bool): True: treat value as signed, False: treat value as unsigned.
        action (aerospike.constant): Action taken if an overflow/underflow occurs.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_SUBTRACT,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_KEY: value,
        SIGN_KEY: sign,
        ACTION_KEY: action
    }


def bit_xor(bin_name: str, bit_offset, bit_size, value_byte_size, value, policy=None):
    """Creates a bit_xor_operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Creates a bit and operation. Server performs bitwise xor with value and bitmap in bin
    at bit_offset for bit_size. Server returns nothing.

    Args:
        bin_name (str): The name of the bin containing the map.
        bit_offset (int): The offset where the bits will start being compared.
        bit_size (int): How many bits of value to xor.
        value_byte_size (int): Length of value in bytes.
        value (bytes/byte array): Value to be used in xor operation.
        policy (dict): The :ref:`bit_policy <aerospike_bit_policies>` dictionary. default: None.

    Returns:
        A dictionary usable in operate or operate_ordered. The format of the dictionary
        should be considered an internal detail, and subject to change.
    """
    return {
        OP_KEY: aerospike.OP_BIT_XOR,
        BIN_KEY: bin_name,
        POLICY_KEY: policy,
        BIT_OFFSET_KEY: bit_offset,
        BIT_SIZE_KEY: bit_size,
        VALUE_BYTE_SIZE_KEY: value_byte_size,
        VALUE_KEY: value
    }



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
Bitwise expressions contain expressions for performing bitwise operations.
Most of these operations are equivalent to the
:mod:`Bitwise Operations API <aerospike_helpers.operations.bitwise_operations>`
for binary data.

Example::

    import aerospike_helpers.expressions as exp
    # Let blob bin "c" == bytearray([3] * 5).
    # Count set bits starting at 3rd byte in bin "c" to get count of 6.
    expr = exp.BitCount(16, 8 * 3, exp.BlobBin("c")).compile()
'''

#from __future__ import annotations
from itertools import chain
from typing import List, Optional, Tuple, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import ResultType
from aerospike_helpers.expressions.resources import _Keys
from aerospike_helpers.expressions.base import BlobBin

########################
# Bit Modify Expressions
########################

TypeBitValue = Union[bytes, bytearray]
TypeBinName = Union[_BaseExpr, str]
TypePolicy = Union[Dict[str, Any], None]


class BitResize(_BaseExpr):
    """Create an expression that performs a bit_resize operation."""
    _op = aerospike.OP_BIT_RESIZE

    def __init__(self, policy: 'TypePolicy', byte_size: int, flags: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                byte_size (int): Number of bytes the resulting blob should occupy.
                flags (int): One or a combination of bit resize flags.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Blob value expression of resized blob bin.

            Example::

                # Blob bin "c" == bytearray([1] * 5).
                # Resize blob bin "c" from the front so that the returned value is bytearray([0] * 5 + [1] * 5).
                expr = exp.BitResize(None, 10, aerospike.BIT_RESIZE_FROM_FRONT, exp.BlobBin("c")).compile()
        """
        self._children= (
            byte_size,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: flags} if flags is not None else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitInsert(_BaseExpr):
    """Create an expression that performs a bit_insert operation."""
    _op = aerospike.OP_BIT_INSERT

    def __init__(self, policy: 'TypePolicy', byte_offset: int, value: 'TypeBitValue', bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                byte_offset (int): Integer byte index of where to insert the value.
                value (TypeBitValue): A bytes value or blob value expression to insert.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob containing the inserted bytes.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # Insert 3 so that returned value is bytearray([1, 3, 1, 1, 1, 1]).
                expr = exp.BitInsert(None, 1, bytearray([3]), exp.BlobBin("c")).compile()
        """
        self._children= (
            byte_offset,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitRemove(_BaseExpr):
    """Create an expression that performs a bit_remove operation."""
    _op = aerospike.OP_BIT_REMOVE

    def __init__(self, policy: 'TypePolicy', byte_offset: int, byte_size: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                byte_offset (int): Byte index of where to start removing from.
                byte_size (int): Number of bytes to remove.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob containing the remaining bytes.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # Remove 1 element so that the returned value is bytearray([1] * 4).
                expr = exp.BitRemove(None, 1, 1, exp.BlobBin("c")).compile()
        """
        self._children= (
            byte_offset,
            byte_size,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitSet(_BaseExpr):
    """Create an expression that performs a bit_set operation."""
    _op = aerospike.OP_BIT_SET

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: 'TypeBitValue', bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start overwriting.
                bit_size (int): Number of bits to overwrite.
                value (TypeBitValue): Bytes value or blob expression containing bytes to write.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob expression with the bits overwritten.

            Example::

                # Let blob bin "c" == bytearray([0] * 5).
                # Set bit at offset 7 with size 1 bits to 1 to make the returned value bytearray([1, 0, 0, 0, 0]).
                expr = exp.BitSet(None, 7, 1, bytearray([255]), exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitOr(_BaseExpr):
    """Create an expression that performs a bit_or operation."""
    _op = aerospike.OP_BIT_OR

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: 'TypeBitValue', bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # bitwise Or `8` with the first byte of blob bin c so that the returned value is bytearray([9, 1, 1, 1, 1]).
                expr = exp.BitOr(None, 0, 8, bytearray([8]), exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitXor(_BaseExpr):
    """Create an expression that performs a bit_xor operation."""
    _op = aerospike.OP_BIT_XOR

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: 'TypeBitValue', bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # bitwise Xor `1` with the first byte of blob bin c so that the returned value is bytearray([0, 1, 1, 1, 1]).
                expr = exp.BitXor(None, 0, 8, bytearray([1]), exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitAnd(_BaseExpr):
    """Create an expression that performs a bit_and operation."""
    _op = aerospike.OP_BIT_AND

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: 'TypeBitValue', bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # bitwise and `0` with the first byte of blob bin c so that the returned value is bytearray([0, 5, 5, 5, 5]).
                expr = exp.BitAnd(None, 0, 8, bytearray([0]), exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitNot(_BaseExpr):
    """Create an expression that performs a bit_not operation."""
    _op = aerospike.OP_BIT_NOT

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([255] * 5).
                # bitwise, not, all of "c" to get bytearray([254] * 5).
                expr = exp.BitNot(None, 0, 40, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitLeftShift(_BaseExpr):
    """Create an expression that performs a bit_lshift operation."""
    _op = aerospike.OP_BIT_LSHIFT

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, shift: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                shift (int): Number of bits to shift by.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # Bit left shift the first byte of bin "c" to get bytearray([8, 1, 1, 1, 1]).
                expr = exp.BitLeftShift(None, 0, 8, 3, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            shift,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitRightShift(_BaseExpr):
    """Create an expression that performs a bit_rshift operation."""
    _op = aerospike.OP_BIT_RSHIFT

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, shift: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                shift (int): Number of bits to shift by.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([8] * 5).
                # Bit left shift the first byte of bin "c" to get bytearray([4, 8, 8, 8, 8]).
                expr = exp.BitRightShift(None, 0, 8, 1, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            shift,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitAdd(_BaseExpr):
    """Create an expression that performs a bit_add operation.
       Note: integers are stored big-endian.
    """
    _op = aerospike.OP_BIT_ADD

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: int, action: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                value (int): Integer value or expression for value to add.
                action (int): An aerospike bit overflow action.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # Bit add the second byte of bin "c" to get bytearray([1, 2, 1, 1, 1])
                expr = exp.BitAdd(None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: action} if action is not None else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitSubtract(_BaseExpr):
    """ Create an expression that performs a bit_subtract operation.
        Note: integers are stored big-endian.
    """
    _op = aerospike.OP_BIT_SUBTRACT

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: int, action: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start operation.
                bit_size (int): Number of bits to be operated on.
                value (int): Integer value or expression for value to add.
                action (int): An aerospike bit overflow action.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: resulting blob with the bits operated on.

            Example::

                # Let blob bin "c" == bytearray([1] * 5).
                # Bit subtract the second byte of bin "c" to get bytearray([1, 0, 1, 1, 1])
                expr = exp.BitSubtract(None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: action} if action is not None else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitSetInt(_BaseExpr):
    """ Create an expression that performs a bit_set_int operation.
        Note: integers are stored big-endian.
    """
    _op = aerospike.OP_BIT_SET_INT

    def __init__(self, policy: 'TypePolicy', bit_offset: int, bit_size: int, value: int, bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
                bit_offset (int): Bit index of where to start writing.
                bit_size (int): Number of bits to overwrite.
                value (int): Integer value or integer expression containing value to write.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Resulting blob expression with the bits overwritten.

            Example::

                # Let blob bin "c" == bytearray([0] * 5).
                # Set bit at offset 7 with size 1 bytes to 1 to make the returned value bytearray([1, 0, 0, 0, 0]).
                expr = exp.BitSetInt(None, 7, 1, 1, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(_ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: policy['bit_write_flags']} if policy is not None and 'bit_write_flags' in policy else {_Keys.VALUE_KEY: 0}),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


######################
# Bit Read Expressions
######################


class BitGet(_BaseExpr):
    """Create an expression that performs a bit_get operation."""
    _op = aerospike.OP_BIT_GET

    def __init__(self, bit_offset: int, bit_size: int, bin: 'TypeBinName'):
        """ Args:
                bit_offset (int): Bit index of where to start reading.
                bit_size (int): Number of bits to get.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Blob, bit_size bits rounded up to the nearest byte size.

            Example::

                # Let blob bin "c" == bytearray([1, 2, 3, 4, 5).
                # Get 2 from bin "c".
                expr = exp.BitGet(8, 8, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitCount(_BaseExpr):
    """Create an expression that performs a bit_count operation."""
    _op = aerospike.OP_BIT_COUNT

    def __init__(self, bit_offset: int, bit_size: int, bin: 'TypeBinName'):
        """ Args:
                bit_offset (int): Bit index of where to start reading.
                bit_size (int): Number of bits to count.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Blob, bit_size bits rounded up to the nearest byte size.

            Example::

                # Let blob bin "c" == bytearray([3] * 5).
                # Count set bits starting at 3rd byte in bin "c" to get count of 6.
                expr = exp.BitCount(16, 8 * 3, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitLeftScan(_BaseExpr):
    """Create an expression that performs a bit_lscan operation."""
    _op = aerospike.OP_BIT_LSCAN

    def __init__(self, bit_offset: int, bit_size: int, value: bool, bin: 'TypeBinName'):
        """ Args:
                bit_offset (int): Bit index of where to start reading.
                bit_size (int): Number of bits to read.
                value bool: Bit value to check for.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Index of the left most bit starting from bit_offset set to value. Returns -1 if not found.

            Example::

                # Let blob bin "c" == bytearray([3] * 5).
                # Scan the first byte of bin "c" for the first bit set to 1. (should get 6)
                expr = exp.BitLeftScan(0, 8, True, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitRightScan(_BaseExpr):
    """Create an expression that performs a bit_rscan operation."""
    _op = aerospike.OP_BIT_RSCAN

    def __init__(self, bit_offset: int, bit_size: int, value: bool, bin: 'TypeBinName'):
        """ Args:
                bit_offset (int): Bit index of where to start reading.
                bit_size (int): Number of bits to read.
                value bool: Bit value to check for.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Index of the right most bit starting from bit_offset set to value. Returns -1 if not found.

            Example::

                # Let blob bin "c" == bytearray([3] * 5).
                # Scan the first byte of bin "c" for the right most bit set to 1. (should get 7)
                expr = exp.BitRightScan(0, 8, True, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            value,
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )


class BitGetInt(_BaseExpr):
    """Create an expression that performs a bit_get_int operation."""
    _op = aerospike.OP_BIT_GET_INT

    def __init__(self, bit_offset: int, bit_size: int, sign: bool, bin: 'TypeBinName'):
        """ Args:
                bit_offset (int): Bit index of where to start reading.
                bit_size (int): Number of bits to get.
                sign bool: True for signed, False for unsigned.
                bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

            :return: Integer expression.

            Example::

                # Let blob bin "c" == bytearray([1, 2, 3, 4, 5).
                # Get 2 as an integer from bin "c".
                expr = exp.BitGetInt(8, 8, True, exp.BlobBin("c")).compile()
        """
        self._children= (
            bit_offset,
            bit_size,
            1 if sign else 0,
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin)
        )



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
The expressions base module provide expressions for
 * declaring variables, using variables, and control-flow
 * comparison operators
 * applying logical operators to one or more 'boolean expressions'
 * returning the value of (in-memory) record metadata
 * returning the value from storage, such as bin data or the record's key


Example::

    import aerospike_helpers.expressions.base as exp
    # See if integer bin "bin_name" contains a value equal to 10.
    expr = exp.Eq(exp.IntBin("bin_name"), 10).compile()
'''

#from __future__ import annotations
from itertools import chain
from typing import List, Optional, Tuple, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import ResultType
from aerospike_helpers.expressions.resources import _Keys

TypeComparisonArg = Union[_BaseExpr, Any]
TypeGeo = Union[_BaseExpr, aerospike.GeoJSON]


###################
# Value Expressions
###################


class Unknown(_BaseExpr):
    """ Create an 'Unknown' expression, which allows an operation expression
    ('read expression' or 'write expression') to be aborted.

    This expression returns a special 'unknown' trilean value which, when
    returned by an operation expression, will result in an error code 26
    :class:`~aerospike.exception.OpNotApplicable`. These failures can be ignored with the policy flags
    :class:`aerospike.EXP_READ_EVAL_NO_FAIL` for read expressions and
    :class:`aerospike.EXP_WRITE_EVAL_NO_FAIL` for write expressions.
    This would then allow subsequent operations in the transaction to proceed.

    This expression is only useful from a
    :class:`~aerospike_helpers.expressions.base.Cond` conditional expression within
    an operation expression, and should be avoided in filter-expressions, where
    it might trigger an undesired move into the storage-data phase.

    If a test-expression within the ``Cond`` yields the special 'unknown' trilean
    value, then the ``Cond`` will also immediately yield the 'unknown' value and
    further test-expressions will not be evaluated.

    Note that this special 'unknown' trilean value is the same value returned
    by any failed expression.
    """
    _op = _ExprOp.UNKNOWN

    def __init__(self):
        """ :return: (unknown value)

            Example::

                # If IntBin("balance") >= 50, get "balance" + 50.
                # Otherwise, fail the expression via Unknown().
                # This sort of expression is useful with expression operations
                # expression_read() and expression_write().
                exp.Let(exp.Def("bal", exp.IntBin("balance")),
                    exp.Cond(
                        exp.GE(exp.Var("bal"), 50),
                            exp.Add(exp.Var("bal"), 50),
                        exp.Unknown())
                )
        """
        super().__init__()


########################
# Record Key Expressions
########################


class _Key(_BaseExpr):
    _op = _ExprOp.REC_KEY


class KeyInt(_Key):
    """ Create an expression that returns the key as an integer. Returns the unknown-value if
        the key is not an integer.
    """
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): Integer value of the key if the key is an integer.

            Example::

                # Integer record key >= 10000.
                expr = exp.GE(KeyInt(), 10000).compile()
        """
        super().__init__()


class KeyStr(_Key):
    """ Create an expression that returns the key as a string. Returns the unknown-value if
        the key is not a string.
    """
    _rt = ResultType.STRING

    def __init__(self):
        """ :return: (string value): string value of the key if the key is an string.

            Example::

                # string record key == "aaa".
                expr = exp.Eq(exp.KeyStr(), "aaa").compile()
        """ 
        super().__init__()


class KeyBlob(_Key):
    """ Create an expression that returns the key as a blob. Returns the unknown-value if
        the key is not a blob.
    """
    _rt = ResultType.BLOB

    def __init__(self):
        """ :return: (blob value): Blob value of the key if the key is a blob.

            Example::

                # blob record key <= bytearray([0x65, 0x65]).
                expr = exp.GE(exp.KeyBlob(), bytearray([0x65, 0x65])).compile()
        """ 
        super().__init__()


class KeyExists(_BaseExpr):
    """ Create an expression that returns if the primary key is stored in the record storage
        data as a boolean expression. This would occur on record write, when write policies set the `key` field to
        :class:`aerospike.POLICY_KEY_SEND`.
    """
    _op = _ExprOp.META_KEY_EXISTS
    _rt = ResultType.BOOLEAN

    def __init__(self):
        """ :return: (boolean value): True if the record has a stored key, false otherwise.

            Example::

                # Key exists in record meta data.
                expr = exp.KeyExists().compile()
        """ 
        super().__init__()


#################
# Bin Expressions
#################


class BoolBin(_BaseExpr):
    """ Create an expression that returns a bin as a boolean. Returns the unknown-value
        if the bin is not a boolean.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.BOOLEAN

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (boolean bin)

            Example::

                # Boolean bin "a" is True.
                expr = exp.BoolBin("a").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class IntBin(_BaseExpr):
    """ Create an expression that returns a bin as an integer. Returns the unknown-value
        if the bin is not an integer.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.INTEGER

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (integer bin)

            Example::

                # Integer bin "a" == 200.
                expr = exp.Eq(exp.IntBin("a"), 200).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class StrBin(_BaseExpr):
    """ Create an expression that returns a bin as a string. Returns the unknown-value
        if the bin is not a string.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.STRING

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (string bin)

            Example::

                # String bin "a" == "xyz".
                expr = exp.Eq(exp.StrBin("a"), "xyz").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class FloatBin(_BaseExpr):
    """ Create an expression that returns a bin as a float. Returns the unknown-value
        if the bin is not a float.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.FLOAT

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (float bin)

            Example::

                # Float bin "a" > 2.71.
                expr = exp.GT(exp.FloatBin("a"), 2.71).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BlobBin(_BaseExpr):
    """ Create an expression that returns a bin as a blob. Returns the unknown-value
        if the bin is not a blob.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.BLOB

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (blob bin)

            Example::

                #. Blob bin "a" == bytearray([0x65, 0x65])
                expr = exp.Eq(exp.BlobBin("a"), bytearray([0x65, 0x65])).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class GeoBin(_BaseExpr):
    """ Create an expression that returns a bin as a geojson. Returns the unknown-value
        if the bin is not a geojson.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.GEOJSON

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (geojson bin)

            Example::

                #GeoJSON bin "a" contained by GeoJSON bin "b".
                expr = exp.CmpGeo(GeoBin("a"), exp.GeoBin("b")).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class ListBin(_BaseExpr):
    """ Create an expression that returns a bin as a list. Returns the unknown-value
        if the bin is not a list.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.LIST

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (list bin)

            Example::

                # List bin "a" contains at least one item with value "abc".
                expr = exp.GT(exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, 
                            ResultType.INTEGER, "abc", ListBin("a")), 
                        0).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class MapBin(_BaseExpr):
    """ Create an expression that returns a bin as a map. Returns the unknown-value
        if the bin is not a map.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.MAP

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (map bin)

            Example::

                # Map bin "a" size > 7.
                expr = exp.GT(exp.MapSize(None, exp.MapBin("a")), 7).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class HLLBin(_BaseExpr):
    """ Create an expression that returns a bin as a HyperLogLog. Returns the unknown-value
        if the bin is not a HyperLogLog.
    """
    _op = _ExprOp.BIN
    _rt = ResultType.HLL

    def __init__(self, bin: str):
        """ Args:
                bin (str): Bin name.

            :return: (HyperLogLog bin)

            Example::

                # Does HLL bin "a" have a hll_count > 1000000.
                expr = exp.GT(exp.HllGetCount(exp.HllBin("a"), 1000000)).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BinExists(_BaseExpr):
    """Create an expression that returns True if bin exists."""
    _op = _ExprOp.BIN_EXISTS
    _rt = ResultType.BOOLEAN

    def __init__(self, bin: str):
        """ Args:
                bin (str): bin name.

            :return: (boolean value): True if bin exists, False otherwise.

            Example::

                #Bin "a" exists in record.
                expr = exp.BinExists("a").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BinType(_BaseExpr):
    """ Create an expression that returns the type of a bin
        as one of the aerospike :ref:`bin types <aerospike_bin_types>`
    """
    _op = _ExprOp.BIN_TYPE
    _rt = ResultType.INTEGER

    def __init__(self, bin: str):
        """ Args:
                bin (str): bin name.

            :return: (integer value): returns the bin type.

            Example::

                # bin "a" == type string.
                expr = exp.Eq(exp.BinType("a"), aerospike.AS_BYTES_STRING).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


####################
# Record Expressions
####################


class SetName(_BaseExpr):
    """ Create an expression that returns record set name string.
        This expression usually evaluates quickly because record
        meta data is cached in memory.
    """
    _op = _ExprOp.META_SET_NAME
    _rt = ResultType.STRING

    def __init__(self):
        """ :return: (string value): Name of the set this record belongs to.

            Example::

                # Record set name == "myset".
                expr = exp.Eq(exp.SetName(), "myset").compile()
        """
        super().__init__()


class DeviceSize(_BaseExpr):
    """ Create an expression that returns record size on disk. If server storage-engine is
        memory, then zero is returned. This expression usually evaluates quickly
        because record meta data is cached in memory.
    """
    _op = _ExprOp.META_DEVICE_SIZE
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): Uncompressed storage size of the record.

            Example::

                # Record device size >= 100 KB.
                expr = exp.GE(exp.DeviceSize(), 100 * 1024).compile()
        """
        super().__init__()


class LastUpdateTime(_BaseExpr):
    """ Create an expression that the returns record last update time expressed as 64 bit
        integer nanoseconds since 1970-01-01 epoch.
    """
    _op = _ExprOp.META_LAST_UPDATE_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): When the record was last updated.

            Example::

                # Record last update time >= 2020-01-15.
                expr = exp.GE(exp.LastUpdateTime(), 1577836800).compile()
        """
        super().__init__()


class SinceUpdateTime(_BaseExpr):
    """ Create an expression that returns milliseconds since the record was last updated.
        This expression usually evaluates quickly because record meta data is cached in memory.
    """
    _op = _ExprOp.META_SINCE_UPDATE_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): Number of milliseconds since last updated.

            Example::

                # Record last updated more than 2 hours ago.
                expr = exp.GT(exp.SinceUpdateTime(), 2 * 60 * 1000).compile()
        """
        super().__init__()    


class VoidTime(_BaseExpr):
    """ Create an expression that returns record expiration time expressed as 64 bit
        integer nanoseconds since 1970-01-01 epoch.
    """
    _op = _ExprOp.META_VOID_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): Expiration time in nanoseconds since 1970-01-01.

            Example::

                # Record expires on 2021-01-01.
                expr = exp.And(
                        exp.GE(exp.VoidTime(), 1609459200),
                        exp.LT(exp.VoidTime(), 1609545600)).compile()
        """
        super().__init__()  


class TTL(_BaseExpr):
    """ Create an expression that returns record expiration time (time to live) in integer
        seconds.
    """
    _op = _ExprOp.META_TTL
    _rt = ResultType.INTEGER

    def __init__(self):
        """ :return: (integer value): Number of seconds till the record will expire,
                                    returns -1 if the record never expires.

            Example::

                # Record expires in less than 1 hour.
                expr = exp.LT(exp.TTL(), 60 * 60).compile()
        """
        super().__init__()  


class IsTombstone(_BaseExpr):
    """ Create an expression that returns if record has been deleted and is still in
        tombstone state. This expression usually evaluates quickly because record
        meta data is cached in memory. NOTE: this is only applicable for XDR filter expressions.
    """
    _op = _ExprOp.META_IS_TOMBSTONE
    _rt = ResultType.BOOLEAN

    def __init__(self):
        """ :return: (boolean value): True if the record is a tombstone, false otherwise.

            Example::

                # Detect deleted records that are in tombstone state.
                expr = exp.IsTombstone().compile()
        """
        super().__init__() 


class DigestMod(_BaseExpr):
    """Create an expression that returns record digest modulo as integer."""
    _op = _ExprOp.META_DIGEST_MOD
    _rt = ResultType.INTEGER

    def __init__(self, mod: int):
        """ Args:
                mod (int): Divisor used to divide the digest to get a remainder.

            :return: (integer value): Value in range 0 and mod (exclusive).

            Example::

                # Records that have digest(key) % 3 == 1.
                expr = exp.Eq(exp.DigestMod(3), 1).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: mod}


########################
# Comparison Expressions
########################


class Eq(_BaseExpr):
    """Create an equals, (==) expression."""
    _op = _ExprOp.EQ

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
            expr0 (TypeComparisonArg): Left argument to `==`.
            expr1 (TypeComparisonArg): Right argument to `==`.

        :return: (boolean value)

        Example::

            # Integer bin "a" == 11
            expr = exp.Eq(exp.IntBin("a"), 11).compile()
        """
        self._children = (expr0, expr1)


class NE(_BaseExpr):
    """Create a not equals (not ==) expressions."""
    _op = _ExprOp.NE

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
                expr0 (TypeComparisonArg): Left argument to `not ==`.
                expr1 (TypeComparisonArg): Right argument to `not ==`.

            :return: (boolean value)

            Example::

                # Integer bin "a" not == 13.
                expr = exp.NE(exp.IntBin("a"), 13).compile()
        """         
        self._children = (expr0, expr1)


class GT(_BaseExpr):
    """Create a greater than (>) expression."""
    _op = _ExprOp.GT

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
                expr0 (TypeComparisonArg): Left argument to `>`.
                expr1 (TypeComparisonArg): Right argument to `>`.

            :return: (boolean value)

            Example::

                # Integer bin "a" > 8.
                expr = exp.GT(exp.IntBin("a"), 8).compile()
        """
        self._children = (expr0, expr1)


class GE(_BaseExpr):
    """Create a greater than or equal to (>=) expression."""
    _op = _ExprOp.GE

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
                expr0 (TypeComparisonArg): Left argument to `>=`.
                expr1 (TypeComparisonArg): Right argument to `>=`.

            :return: (boolean value)

            Example::

                # Integer bin "a" >= 88.
                expr = exp.GE(exp.IntBin("a"), 88).compile()
        """
        self._children = (expr0, expr1)


class LT(_BaseExpr):
    """Create a less than (<) expression."""
    _op = _ExprOp.LT

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
                expr0 (TypeComparisonArg): Left argument to `<`.
                expr1 (TypeComparisonArg): Right argument to `<`.

            :return: (boolean value)

            Example::

                # Integer bin "a" < 1000.
                expr = exp.LT(exp.IntBin("a"), 1000).compile()
        """
        self._children = (expr0, expr1)


class LE(_BaseExpr):
    """Create a less than or equal to (<=) expression."""
    _op = _ExprOp.LE

    def __init__(self, expr0: 'TypeComparisonArg', expr1: 'TypeComparisonArg'):
        """ Args:
                expr0 (TypeComparisonArg): Left argument to `<=`.
                expr1 (TypeComparisonArg): Right argument to `<=`.

            :return: (boolean value)

            Example::

                # Integer bin "a" <= 1.
                expr = exp.LE(exp.IntBin("a"), 1).compile()
        """
        self._children = (expr0, expr1)


class CmpRegex(_BaseExpr):
    """ Create an expression that performs a regex match on a string bin or value expression."""
    _op = _ExprOp.CMP_REGEX

    def __init__(self, options: int, regex_str: str, cmp_str: Union[_BaseExpr, str]):
        """ Args:
                options (int) :ref:`regex_constants`: One of the aerospike regex constants, :ref:`regex_constants`.
                regex_str (str): POSIX regex string.
                cmp_str (Union[_BaseExpr, str]): String expression to compare against.

            :return: (boolean value)

            Example::

                # Select string bin "a" that starts with "prefix" and ends with "suffix".
                # Ignore case and do not match newline.
                expr = exp.CmpRegex(aerospike.REGEX_ICASE | aerospike.REGEX_NEWLINE, "prefix.*suffix", exp.BinStr("a")).compile()
        """
        self._children = (cmp_str,)
        self._fixed = {_Keys.REGEX_OPTIONS_KEY: options, _Keys.VALUE_KEY: regex_str}


class CmpGeo(_BaseExpr):
    """Create a point within region or region contains point expression."""
    _op = _ExprOp.CMP_GEO

    def __init__(self, expr0: 'TypeGeo', expr1: 'TypeGeo'):
        """ Args:
                expr0 (TypeGeo): Left expression in comparrison.
                expr1 (TypeGeo): Right expression in comparrison.

            :return: (boolean value)

            Example::

                # Geo bin "point" is within geo bin "region".
                expr = exp.CmpGeo(GeoBin("point"), exp.GeoBin("region")).compile()
        """
        self._children = (expr0, expr1)


#####################
# Logical Expressions
#####################


class Not(_BaseExpr):
    """Create a "not" (not) operator expression."""
    _op = _ExprOp.NOT

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
                `*exprs` (_BaseExpr): Variable amount of expressions to be negated.

            :return: (boolean value)

            Example::

                # not (a == 0 or a == 10)
                expr = exp.Not(exp.Or(
                            exp.Eq(exp.IntBin("a"), 0),
                            exp.Eq(exp.IntBin("a"), 10))).compile()
        """
        self._children = exprs


class And(_BaseExpr):
    """Create an "and" operator that applies to a variable amount of expressions."""
    _op = _ExprOp.AND

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be ANDed together.

        :return: (boolean value)

        Example::

            # (a > 5 || a == 0) && b < 3
            expr = exp.And(
                    exp.Or(
                      exp.GT(exp.IntBin("a"), 5),
                      exp.Eq(exp.IntBin("a"), 0)),
                    exp.LT(exp.IntBin("b"), 3)).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Or(_BaseExpr):
    """Create an "or" operator that applies to a variable amount of expressions."""
    _op = _ExprOp.OR

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be ORed together.

        :return: (boolean value)

        Example::

            # (a == 0 || b == 0)
            expr = exp.Or(
                    exp.Eq(exp.IntBin("a"), 0),
                    exp.Eq(exp.IntBin("b"), 0)).compile()
        """ 
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Exclusive(_BaseExpr):
    """Create an expression that returns True if only one of the expressions are True."""
    _op = _ExprOp.EXCLUSIVE

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be checked.

        :return: (boolean value)

        Example::

            # exclusive(a == 0, b == 0)
            expr = exp.Exclusive(
                            exp.Eq(exp.IntBin("a"), 0),
                            exp.Eq(exp.IntBin("b"), 0)).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


#######################################
# Flow Control and Variable Expressions
#######################################


class Cond(_BaseExpr):
    """ Conditionally select an expression from a variable number of
        condition/action pairs, followed by a default expression action.

        Takes a set of test-expression/action-expression pairs and evaluates
        each test expression, one at a time. If a test returns ``True``, ``Cond``
        evaluates the corresponding action expression and returns its value,
        after which ``Cond`` doesn't evaluate any of the other tests or
        expressions.  If all tests evaluate to ``False``, the default action
        expression is evaluated and returned.

        ``Cond`` is strictly typed, so all actions-expressions must evaluate to
        the same type or the :class:`~aerospike_helpers.expressions.base.Unknown` expression.

        Requires server version 5.6.0+.
    """
    _op = _ExprOp.COND

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
            `*exprs` (_BaseExpr): bool exp1, action exp1, bool exp2, action exp2, ..., action-default

        :return: (boolean value)

        Example::

            # Apply operator based on type and test if greater than 100.
            expr = exp.GT(
                    exp.Cond(
                        exp.Eq(exp.IntBin("type"), 0),
                            exp.Add(exp.IntBin("val1"), exp.IntBin("val2")),
                        exp.Eq(exp.IntBin("type"), 1),
                            exp.Sub(exp.IntBin("val1"), exp.IntBin("val2")),
                        exp.Eq(exp.IntBin("type"), 2),
                            exp.Mul(exp.IntBin("val1"), exp.IntBin("val2")))
                    100).compile()

        Example::

            # Delete the 'grade' bin if its value is less than 70
            killif = exp.Cond(
                exp.LT(exp.IntBin("grade"), 70), aerospike.null(),
                exp.Unknown()).compile()
            # Write a NIL on grade < 70 to delete the bin
            # or short-circuit out of the operation without raising an exception
            ops = [
                opexp.expression_write("grade", killif,
                aerospike.EXP_WRITE_ALLOW_DELETE | aerospike.EXP_WRITE_EVAL_NO_FAIL),
            ]
            """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Let(_BaseExpr):
    """ Defines variables to be used within the ``Let`` expression's scope. The last
        argument can be any expression and should make use of the defined
        variables. The ``Let`` expression returns the evaluated result of the last
        argument. This expression is useful if you need to reuse the result of a
        complicated or expensive expression.
    """
    _op = _ExprOp.LET

    def __init__(self, *exprs: _BaseExpr):
        """ Args:
            `*exprs` (_BaseExpr): Variable number of :class:`~aerospike_helpers.expressions.base.Def` expressions followed by a scoped expression.

        :return: (result of scoped expression)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                    exp.And(
                        exp.LT(5, exp.Var("x")),
                        exp.LT(exp.Var("x"), 10))).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Def(_BaseExpr):
    """ Assign variable to an expression that can be accessed later with :class:`~aerospike_helpers.expressions.base.Var`.
        Requires server version 5.6.0+.
    """
    _op = _ExprOp.DEF

    def __init__(self, var_name: str, expr: _BaseExpr):
        """ Args:
            `var_name` (str): Variable name.
            `expr` (_BaseExpr): Variable is set to result of this expression.

        :return: (a variabe name expression pair)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                    exp.And(
                        exp.LT(5, exp.Var("x")),
                        exp.LT(exp.Var("x"), 10))).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: var_name}
        self._children = (expr,)


class Var(_BaseExpr):
    """ Retrieve expression value from a variable previously defined with :class:`~aerospike_helpers.expressions.base.Def`.
        Requires server version 5.6.0+.
    """
    _op = _ExprOp.VAR

    def __init__(self, var_name: str):
        """ Args:
            `var_name` (str): Variable name.

        :return: (value stored in variable)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                    exp.And(
                        exp.LT(5, exp.Var("x")),
                        exp.LT(exp.Var("x"), 10))).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: var_name}
